--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

-------------------------------------- LOCALISATIONS -------------------------------------------------
local TEXTS = {
	NoConfig = "Configuration is not available!",
	Range = "Range: ",
	Unknown = "Range?",
}

--[[
local locale = GetLocale()
if locale == "frFR" then
	TEXTS = {
		NoConfig = "French for: Configuration is not available!",
--		Range = "French for: Range: ",
--		Unknown = "French for: Range?",
	}
end
]]--

------------------------------------------------------------------------------------------------------

local floor, ceil = math.floor, math.ceil
local sub = string.sub

local LibRange = LibStub("LibRangeCheck-3.0", true)

local PlayerClass, TargetingName
local Bear, Cat, RangedText, Show, OOC = 1, 2, 1, 1, 2
local Events = {}

NS.Force = { default=1, profile=2 } 
NS.AltMana, NS.HasPet  = 1, 2

--local DebugPrint = FIZZLEDEBUG and FIZZLEDEBUG(AddonName) or function() end
--------------------------------------------------
--	Classic Check for Retail global		--
--------------------------------------------------
local Alt_Mana_Bar_Pair_Display_Info = ALT_MANA_BAR_PAIR_DISPLAY_INFO
if not Alt_Mana_Bar_Pair_Display_Info then
	Alt_Mana_Bar_Pair_Display_Info = {
		DRUID = {
			[Enum.PowerType.Rage] = true,
			[Enum.PowerType.Energy] = true,
		},
	}
end

local function OnEvent(self, event, ...)
	Events[event](self, ...)
end

local function UpdateFont(parent, setpoint, toopoint, x, y, justification, font, size, colourtable, create, texttype)
	local text
	local label = "$parent_"
	if texttype == RangedText then
		label = label.."RANGE_"
	end
	label = label.."TEXT"
	if create then
		text = parent:CreateFontString(label, "BACKGROUND")
		text.parent = parent
	else
		text = parent.text or parent
	end
	text:ClearAllPoints()
	local parent
	if texttype and texttype ~= RangedText then
		parent = texttype
	else
		parent = text.parent
	end
	text:SetPoint(setpoint, parent, toopoint, x, y)
	text:SetFont(font, size)
	text:SetJustifyH(justification)
	text:SetTextColor(colourtable.r, colourtable.g, colourtable.b)
	return text
end

local function Divide(number, divisor)
	if number == 0 then
		return(0)
	end
	if divisor == 0 then
		return number
	end
	return number/divisor
end

local function GetLevel(current, max)
	local percent = 0
	if max > 0 then 
		percent = Divide(current, max)
	end
	return percent
end

local function OnUpdate(self, elapsed)
	if self.updatetimer == nil then
		self.updatetimer = 0
		self.totalelapsed = 0
	end
	self.updatetimer = self.updatetimer - elapsed
	self.totalelapsed = self.totalelapsed + elapsed
	if (self.updatetimer > 0) then return end
	self.updatetimer = NS.Settings.UpdateSpeed
	local min, max = LibRange:getRange("target")
	local rangetext = TEXTS.Range or ""
	if min then
		if max then
			self.rangetext:SetText(rangetext..min.." - "..max)
		else
			self.rangetext:SetText(rangetext..min.."+")
		end
	else
		self.rangetext:SetText(TEXTS.Unknown)
	end
end

local function UpdateAlpha(self, unit, visibility)
	local bar = self.Units[unit].Frames
	local bgalpha, alpha = 0, 0
	if visibility == Show then
		bgalpha, alpha = NS.Settings.BackgroundAlpha, NS.Settings.Alpha
	elseif visibility == OOC then
		bgalpha, alpha = NS.Settings.BackgroundAlphaOOC, NS.Settings.AlphaOOC
	end
	bar.bghb:SetAlpha(bgalpha)
	bar.fghb:SetAlpha(alpha)
	if alpha > 0 and bar.bgalt then
		bar.bgalt:SetAlpha(1) -- alpha of parent
		bar.fgalt:SetAlpha(1) -- alpha of parent
	end
end

local function UpdateVisibility(self)
	local visible, rangevisible
	local targetexists = UnitExists("target")
	if NS.Settings.HideForFriendlies then
		if targetexists and self.rangetext.show then
			rangevisible = true
		end
		targetexists = UnitCanAttack("player", "target")
	end
	if targetexists then
		UpdateAlpha(self, "target", Show)
		UpdateAlpha(self, "player", Show)
		visible = true
	else
		UpdateAlpha(self, "target")
		if NS.Settings.HideAtMaxHealth and self.health == self.healthmax then
			UpdateAlpha(self, "player")
		else
			UpdateAlpha(self, "player", OOC)
			visible = true
		end
	end
	if visible and not self.AltStatus then
		self.Units.player.Frames.bgalt:SetAlpha(0)
		self.Units.player.Frames.fgalt:SetAlpha(0)
	end
	if NS.ConfigOpen then
		if NS.ConfigAlphaNotarget then
			UpdateAlpha(self, "target", OOC)
			UpdateAlpha(self, "player", OOC)
		else
			UpdateAlpha(self, "target", Show)
			UpdateAlpha(self, "player", Show)
		end
		rangevisible = true
		visible = true
	end
	if self.rangetext.show then
		if rangevisible then
			self.rangetext:SetAlpha(NS.Settings.Alpha)
		else
			self.rangetext:SetAlpha(0)
		end
	end
	if visible or rangevisible then
		self:SetScript("OnUpdate", OnUpdate)
	else
		self:SetScript("OnUpdate", nil)
	end
	self.visible = visible
end

local function UpdateBarValue(frame, value)
	local height = frame.height * value
	frame:SetHeight(height)
	frame.texture:SetHeight(height)
	local tc = 1 - Divide(height, frame.origheight)
	if not frame.flip then
		frame.texture:SetTexCoord(0, 1, tc, 1)
	else
		frame.texture:SetTexCoord(1, 0, tc, 1)
	end
end

local function UpdateBar(bar, val, maxval, textaspercent, showmax, unit)
	local Percent
	if maxval == 0 or maxval < val then
		Percent = 0
	else
		Percent = GetLevel(val, maxval)
	end
	if NS.ConfigLevel and NS.ConfigLevel > -1 then
		val = NS.ConfigLevel
		if maxval == 0 then
			maxval = 100
		end
		Percent = NS.ConfigLevel
		val = maxval/100 * (NS.ConfigLevel * 100)
	end
	UpdateBarValue(bar, Percent)
	if NS.Settings.HideNumbers then return end
	local Text
	if Percent > 0 then
		if textaspercent then
			--Text = Round(Divide(val, maxval) * 100, 0) .. "%" 
			Text = ceil((val / maxval) * 100) .. "%" 
		else
			Text = BreakUpLargeNumbers(val)
		end
		if showmax then
			Text = Text.. "\n" .. BreakUpLargeNumbers(maxval)
		end
	end
	bar.text:SetText(Text)
end

local function UpdateHealthBar(self, unit, force)
	if unit ~= "player" and unit ~= "target" and unit ~= "pet" then return end
	local UnitBar
	if unit == "pet" then
		UnitBar = self.Units.player.Frames.fgalt
	else
		UnitBar = self.Units[unit].Frames.fghb
	end
	if unit == "target" then
		if not UnitPlayerControlled("target") and UnitIsTapDenied("target") then
			if not UnitBar.targetNotMine then
				UnitBar.targetNotMine = true
				local colour = NS.Settings.Units.target.fg.hb.tapped
				UnitBar.texture:SetVertexColor(colour.r, colour.g, colour.b)
			end
		else
			if UnitBar.targetNotMine then
				UnitBar.targetNotMine = nil
				local colour = NS.Settings.Units.target.fg.hb.colour
				UnitBar.texture:SetVertexColor(colour.r, colour.g, colour.b)
			end
		end
	end
	local health, healthmax = UnitHealth(unit), UnitHealthMax(unit)
	if NS.Settings.HideAtMaxHealth and unit == "player" then
		self.health, self.healthmax = health, healthmax
		if (health == healthmax and self.visible) or (health < healthmax and not self.visible) then
			UpdateVisibility(self)
		end
	end
	if force or UnitBar:IsShown() then
		UpdateBar(UnitBar, health, healthmax, NS.Settings.HealthAsPercent, NS.Settings.HealthShowMax)
	end
end 

local function UpdatePowerBar(self, unit, force)
	if unit ~= "player" and unit ~= "target" then return end
	local powertype, t = UnitPowerType(unit)
	if not self.Units[unit].Frames.fghb:IsShown() and not force then return end
	local power, powermax = UnitPower(unit, powertype), UnitPowerMax(unit, powertype)
	UpdateBar(self.Units[unit].Frames.fgpb, power, powermax, NS.Settings.PowerAsPercent, NS.Settings.PowerShowMax, unit)
	if unit == "player" and self.AltStatus == NS.AltMana then
		power = UnitPower(unit, Enum.PowerType.Mana)
		powermax = UnitPowerMax(unit, Enum.PowerType.Mana)
		UpdateBar(self.Units[unit].Frames.fgalt, power, powermax, NS.Settings.PowerAsPercent, NS.Settings.PowerShowMax)
	end
end

local function UpdateAltBarStatus(self)
	self.AltStatus = nil
--	if UnitPowerType("player") ~= 0 and UnitPowerMax("player", UnitPowerType("player")) > 0 and  UnitPowerMax("player", 0) > 0 then -- main power not mana and two max. powers greater than zero
	if Alt_Mana_Bar_Pair_Display_Info[PlayerClass] and Alt_Mana_Bar_Pair_Display_Info[PlayerClass][UnitPowerType("player")] -- THIS SHOULDN'T NEED A LOOKUP BLIZ!!! Stupid Monks
		or PlayerClass == "DRUID" and UnitPowerType("player") ~= Enum.PowerType.Mana then 
			self.AltStatus = NS.AltMana
	end
	if UnitExists("pet") then  -- Pet health takes priority over NS.AltMana user
		self.AltStatus = NS.HasPet
	end
	if NS.ConfigOpen then -- Config takes priority over everything
		self.AltStatus = NS.ConfigAltBarState
	end
	local altbg = self.Units.player.Frames.bgalt
	local altfg = self.Units.player.Frames.fgalt
	local playerbgpb = self.Units.player.Frames.bgpb
	local playerfgpb = self.Units.player.Frames.fgpb
	if self.AltStatus then
		playerbgpb.texture:SetTexture(NS.Settings.TexturePath..playerbgpb.alttexture)
		playerfgpb.texture:SetTexture(NS.Settings.TexturePath..playerfgpb.alttexture)
		if self.AltStatus == NS.AltMana then
			altbg.texture:SetVertexColor(altbg.colour.r, altbg.colour.g, altbg.colour.b)
			altfg.texture:SetVertexColor(altfg.colour.r, altfg.colour.g, altfg.colour.b)
		else
			altbg.texture:SetVertexColor(altbg.petcolour.r, altbg.petcolour.g, altbg.petcolour.b)
			altfg.texture:SetVertexColor(altfg.petcolour.r, altfg.petcolour.g, altfg.petcolour.b)
		end
	else
		playerbgpb.texture:SetTexture(NS.Settings.TexturePath..playerbgpb.stdtexture)
		playerfgpb.texture:SetTexture(NS.Settings.TexturePath..playerfgpb.stdtexture)
	end
	UpdateVisibility(self)
end

local function UpdateAll(self)
	UpdateAltBarStatus(self) --UpdateState(self)
	for unit, utable in pairs(self.Units) do
		for bar, btable in pairs(utable.Frames) do
			if btable.layer == "bg" then
				UpdateBarValue(btable, 1)
			else
				if btable.bar == "pb" then
					UpdatePowerBar(self, unit, true)
				elseif btable.bar == "hb" then
					UpdateHealthBar(self, unit, true)
				elseif self.AltStatus then
					if self.AltStatus == NS.AltMana then
						UpdatePowerBar(self, "player", true)
					else
						UpdateHealthBar(self, "pet", true)
					end
				end
			end
		end
	end
end

local function CreateHUDFrame(parent, unit, layer, bar, settings)
	local f = CreateFrame("Frame", AddonName.."_".. unit.."_"..layer.."_"..bar, parent)
	f:SetMovable(false)
	f:EnableMouse(false)
	f:EnableMouseWheel(false)
	f:SetWidth(settings.w or 128)
	f:SetHeight(settings.h or 256)
	f.origheight = settings.h or 256
	f:SetFrameStrata("BACKGROUND")
	local framelevel = 2
	if layer == "bg" then
		framelevel = 1
	end
	f:SetFrameLevel(framelevel)
	f.texture = f:CreateTexture("$parent_Texture","BACKGROUND")
	f.texture:ClearAllPoints()
	f.texture:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
	if settings.ch then
		f:SetHeight(settings.ch)
	end
	f.height = settings.ch or settings.h
	return f
end

local function CreateHUD(self)
	if self.Units then return end
	self.Units = {}
	local barSort = { "hb", "pb", "alt" }
	local font
	for unit, utable in pairs(NS.Settings.Units) do
		local uf = CreateFrame("Frame", "$parent_"..unit, self)
		uf.Frames = {}
		self.Units[unit] = uf
		uf:SetAllPoints()
		local parent = self
		for layer, ltable in pairs(utable) do
			for id, bar in ipairs(barSort) do
				local btable = ltable[bar]
				if not btable then break end
				local b = CreateHUDFrame(self.Units[unit].Frames[layer.."hb"] or uf, unit, layer, bar, btable)
				if btable.font then
					font = btable.font
					b.text = UpdateFont(b, font.point, font.topoint, font.x, font.y, font.justify, NS.Settings.FontPath..font.font, font.size, font.colour, true)
				end
				b.layer = layer
				b.bar = bar
				self.Units[unit].Frames[layer..bar] = b
			end
		end
	end
	font = NS.Settings.Range.font
	self.rangetext = UpdateFont(self.Units.target.Frames.bghb, font.point, font.topoint, font.x, font.y, font.justify, NS.Settings.FontPath..font.font, font.size, font.colour, true, RangedText)
end

------------------------------- Get it started ----------------------------------------
local f = CreateFrame("Frame", "FAUXMAZZLE_HUD", UIParent)
f:SetSize(10, 10)
f:SetPoint("CENTER", UIParent)
f:Hide()
f:RegisterEvent("PLAYER_LOGIN")
f:SetScript("OnEvent", OnEvent)

function f:SetEnable(enabled)
	if enabled then
		NS:InitSettings()
		CreateHUD(self)
		self.LastPower = 0
		self.updatespeed = NS.Settings.UpdateSpeed
		self.updatetimer = 0
		self.totalelapsed = 0
		NS:UpdateSettings() 
		for k,_ in pairs(Events) do
			self:RegisterEvent(k)
		end
		self:UnregisterEvent("UNIT_FLAGS")
		f:SetScript("OnEvent", OnEvent)
		self:Show()
	else
		self:SetScript("OnUpdate", nil)
		self:SetScript("OnEvent", nil)
		self:UnregisterAllEvents()
		self:Hide()
	end
end

---------------------------------------------------------------------------------------
function FauxMazzleHUD_SetProfile()
	NS:InitSettings(NS.Force.profile)
	if NS.InitAll then
		NS:InitAll()
	end
	NS:UpdateSettings()
end

function NS:InitSettings(force)
	if NS.Settings and not force then return end
	if not FAUXMAZZLEHUD then
		FAUXMAZZLEHUD =  { Profiles={}, ProfileKeys={}, Enabled = true }
	end
	if not FAUXMAZZLEHUD.Profiles.Default or force == NS.Force.default then
		local Defaults = NS:LoadDefaults()
		FAUXMAZZLEHUD.Profiles.Default = {}
		NS:Copy_Table(Defaults, FAUXMAZZLEHUD.Profiles.Default)
	end
	NS.Defaults = FAUXMAZZLEHUD.Profiles.Default
	NS.Realm = GetRealmName() 
	NS.Player = UnitName("player")
	if not FAUXMAZZLEHUD.ProfileKeys[NS.Realm] then
		FAUXMAZZLEHUD.ProfileKeys[NS.Realm] = {}
	end
	if not FAUXMAZZLEHUD.ProfileKeys[NS.Realm][NS.Player] or not FAUXMAZZLEHUD.Profiles[FAUXMAZZLEHUD.ProfileKeys[NS.Realm][NS.Player]] then
		FAUXMAZZLEHUD.ProfileKeys[NS.Realm][NS.Player] = "Default"
	end
	NS.Profile = FAUXMAZZLEHUD.ProfileKeys[NS.Realm][NS.Player]
	NS.Settings = FAUXMAZZLEHUD.Profiles[NS.Profile]
	if TargetingName then return end
	TargetingName = GetUnitName("player", true) -- to be use if target gets targettarget insert and show if GetUnitName("targettarget", true) ~= TargetingName (player)
	PlayerClass = select(2, UnitClass("player"))
end

function NS:UpdateSettings()
	local self = FAUXMAZZLE_HUD
	for unit, utable in pairs(NS.Settings.Units) do
		for layer, ltable in pairs(utable) do
			for bar, btable in pairs(ltable) do
				local settings = btable
				local b = self.Units[unit].Frames[layer..bar]
				b.flip = settings.flip
				if bar == "pb" and settings.alttexture then 
					b.alttexture = settings.alttexture
					b.stdtexture = settings.texture
				end
				if bar == "alt" then
					b.colour = settings.colour
					b.petcolour = settings.petcolour
				end
				b.texture:SetTexture(NS.Settings.TexturePath..settings.texture)
				b.texture:SetTexCoord(settings.texcoord[1], settings.texcoord[2], settings.texcoord[3], settings.texcoord[4])
				b.texture:SetVertexColor(settings.colour.r, settings.colour.g, settings.colour.b)
				local anchor
				if layer == "bg" and bar == "hb" then
					anchor = UIParent
				else
					anchor = self.Units[unit].Frames.bghb
				end
				b:ClearAllPoints()
				b:SetPoint(settings.point, anchor, settings.point, settings.osx, settings.osy)
				local font = settings.font
				if font then
					local relative = self.Units[unit].Frames["bg"..bar]
					UpdateFont(b.text, font.point, font.topoint, font.x, font.y, font.justify, NS.Settings.FontPath..font.font, font.size, font.colour, nil, relative)
				end
				if layer == "fg" and b.text then
					if NS.Settings.HideNumbers then
						b.text:Hide()
					else
						b.text:Show()
					end
				end
			end
		end
	end
	if NS.Settings.HideForFriendlies then
		self:RegisterUnitEvent("UNIT_FLAGS", "target")
	else
		self:UnregisterEvent("UNIT_FLAGS")
	end
	local font = NS.Settings.Range.font
	UpdateFont(self.rangetext, font.point, font.topoint, font.x, font.y, font.justify, NS.Settings.FontPath..font.font, font.size, font.colour, nil, true)
	self.rangetext:ClearAllPoints()
	self.rangetext:SetPoint(font.point, self, font.topoint, font.x, font.y)
	if NS.Settings.Range.ShowForFriendlies then
		self.rangetext:SetParent(self)
		self.rangetext.show = true
	else
		self.rangetext:SetParent(self.Units.target.Frames.bghb)
		self.rangetext.show = nil
	end
	self:SetScale(NS.Settings.Scale)
	self.updatespeed = NS.Settings.UpdateSpeed
	self.updatetimer = 0
	self.totalelapsed = 0
	UpdateAll(self)
end

function NS:Copy_Table(src, dest)
	for index, value in pairs(src) do
		if type(value) == "table" then
			if not dest[index] or type(dest[index]) ~= "table" then
				dest[index] = {}
			end
			NS:Copy_Table(value, dest[index])
		else
			dest[index] = value
		end
	end
end

---------------------------  EVENT Functions -------------------------------------------------------------------
function Events.PLAYER_LOGIN(self, ...)
	if FAUXMAZZLEHUD then
		self:SetEnable(FAUXMAZZLEHUD.Enabled)
	else
		self:SetEnable(true)	
	end
	Events.UNIT_DISPLAYPOWER(self)
end

function Events.UNIT_FLAGS(self, unit)
	UpdateAll(self)
end

function Events.UNIT_PET(self, ...)
	local arg1 = ...
	if  arg1 == "player" then
		UpdateAltBarStatus(self)
		UpdateHealthBar(self, "pet", true)
	end
end

function Events.UNIT_HEALTH(self, ...)
	UpdateHealthBar(self, ...)
end
Events.UNIT_MAXHEALTH = Events.UNIT_HEALTH
if WOW_PROJECT_ID == WOW_PROJECT_CLASSIC then
	Events.UNIT_HEALTH_FREQUENT = Events.UNIT_HEALTH
end

function Events.UNIT_POWER_FREQUENT(self, ...)
	UpdatePowerBar(self, ...)
end
Events.UNIT_MAXPOWER = Events.UNIT_POWER_FREQUENT

function Events.PLAYER_TARGET_CHANGED(self, unit)
	if unit and unit ~= "player" and unit ~= "target" then return end
	UpdateAll(self)
end

Events.UNIT_FACTION = Events.PLAYER_TARGET_CHANGED
	
function Events.PLAYER_ENTERING_WORLD(self, isInitialLogin, isReloadingUi)
	Events.PLAYER_ENTERING_WORLD = Events.PLAYER_TARGET_CHANGED
	UpdateAll(self)
end

function Events.UNIT_DISPLAYPOWER(self, ...) -- Test event 8.1.2 addition
	UpdateAltBarStatus(self)
	UpdatePowerBar(self, "player")
end

---------------------------- Slash Handler -------------------------------------------------------------------
SLASH_FAUZMAZZLHUD1 = "/fmh"
SLASH_FAUZMAZZLHUD2 = "/fmhud"
SLASH_FAUZMAZZLHUD3 = "/hud"
SlashCmdList["FAUZMAZZLHUD"] = function(msg)
	if not NS.CreateCheckBox then
		print(TEXTS.NoConfig)
	else
		local toggle = true
		if strupper(msg) == "RESET" then
			FAUXMAZZLEHUD.ConfigPos = { xoffset=0, yoffset=0 }
			local f = FauxMazzleHUD_ConfigContainer
			if f then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", UIParent, "TOPLEFT")
				if f:IsShown() then
					toggle = nil
				end
			end
		end
		if toggle then
			if FAUXMAZZLE_HUD.ExternalOptions then
				FAUXMAZZLE_HUD.ExternalOptions()
			else
				FAUXMAZZLE_HUD.ToggleOptions()
			end
		end
	end
end
